local isWoW3dot1 = select(2,GetBuildInfo()) >= "9614"
if (isWoW3dot1) then
	return				-- Engine does nothing in 3.1
end

local Utopia = LibStub("AceAddon-3.0"):GetAddon("Utopia", true)
if (not Utopia) then
	return
end
Utopia:UpdateVersion("$Revision: 44 $")

local new, del, copy, deepDel = Utopia.new, Utopia.del, Utopia.copy, Utopia.deepDel
local Gradient, UnitFullName, ClassColour, SmoothColour = Utopia.Gradient, Utopia.UnitFullName, Utopia.ClassColour, Utopia.SmoothColour
local rotate = Utopia.rotate
local band = bit.band
local savageCombat = GetSpellInfo(58413)
local bloodFrenzy = GetSpellInfo(29859)
local db, playerGUID, playerClass

----------------- BUG FIX ----- Remove this when WoW includes Savage Combat/Blood Frenzy in combat log --------------
local SChaxCache = setmetatable({},{
	__index = function(self,key)
		local val, class = UnitClass(key)
		if (class == "ROGUE") then
			val = Utopia:HasTalent(key, GetSpellInfo(58413))	-- Savage Combat
		elseif (class == "WARRIOR") then
			val = Utopia:HasTalent(key, GetSpellInfo(29859))	-- Blood Frenzy
		end
		self[key] = val or false
		return val
	end
})
--------------END BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------

local engine = CreateFrame("Frame", "UtopiaDebuffActivity")
--@debug@
E = engine
function engine:Print(...)
	ChatFrame1:AddMessage("|cFFFFFF80Engine: |r"..format(...))
end
--@no-debug@

function engine:SetDB(ndb)
	db = ndb
end

engine:SetScript("OnEvent", function(self, event, ...)
	local f = engine[event]
	if (f) then
		f(self, ...)
	else
		error("No handler for "..event)
	end
end)

-- IsBlacklistedMobID
local IsBlacklistedMobID
do
	local blackListMobIDs = {
		[19833] = true,		-- Hunter, Snake Trap snake
		[19921] = true,		-- Hunter, Snake Trap snake
	}

	function IsBlacklistedMobID(guid)
		return blackListMobIDs[tonumber(strsub(guid, -12,-7),16)] == true
	end	
end

-- Start
function engine:Start()
	playerGUID = UnitGUID("player")
	playerClass = select(2, UnitClass("player"))
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:RegisterEvent("UNIT_FLAGS")
	self:RegisterEvent("RAID_ROSTER_UPDATE")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED")
	self:RegisterEvent("UNIT_NAME_UPDATE")
	self:RegisterEvent("UNIT_PET")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("PLAYER_FOCUS_CHANGED")
	self:RegisterEvent("PLAYER_ENTERING_WORLD")

	self.mobs = new()
	self.players = new()
	self:ScanRoster()
	self:ResetInCombatForAll()
	self:CheckTargetBuffs("player")
end

-- Stop
function engine:Stop()
	self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:UnregisterEvent("UNIT_FLAGS")
	self:UnregisterEvent("RAID_ROSTER_UPDATE")
	self:UnregisterEvent("PARTY_MEMBERS_CHANGED")
	self:UnregisterEvent("UNIT_NAME_UPDATE")
	self:UnregisterEvent("UNIT_PET")
	self:UnregisterEvent("UNIT_AURA")
	self:UnregisterEvent("PLAYER_TARGET_CHANGED")
	self:UnregisterEvent("PLAYER_FOCUS_CHANGED")

	self.mobs = deepDel(self.mobs)
	self.players = deepDel(self.players)
	self.combat = deepDel(self.combat)
end

local function checkRosterUpdate(self, elapsed)
	if (self.rosterUpdate) then
		self.rosterUpdate = nil
		self:SetScript("OnUpdate", nil)
		self:ScanRoster()
	end
end

function engine:RAID_ROSTER_UPDATE()
	self.rosterUpdate = true
	self:SetScript("OnUpdate", checkRosterUpdate)
end
engine.PARTY_MEMBERS_CHANGED = engine.RAID_ROSTER_UPDATE

-- PLAYER_TARGET_CHANGED
function engine:PLAYER_TARGET_CHANGED()
	self.targetGUID = UnitGUID("target")
	self:CheckTargetDebuffs("target")
end

-- PLAYER_FOCUS_CHANGED
function engine:PLAYER_FOCUS_CHANGED()
	self.focusGUID = UnitGUID("focus")
	self:CheckTargetDebuffs("focus")
end

-- PLAYER_ENTERING_WORLD
function engine:PLAYER_ENTERING_WORLD()
	self.createdTotems = del(self.createdTotems)
	self.createdTotemGUIDs = del(self.createdTotemGUIDs)
end

-- CheckTargetDebuffs
function engine:CheckTargetDebuffs(target)
	-- Validate the target and load up debuffs if it has any
	-- Mark it as temporary and dump them again if we drop target out of combat

	for guid,mob in pairs(self.mobs) do
		if (mob.temp) then
			for debuffName,debuff in pairs(mob.debuffs) do
				if (debuff.temp) then
					--self:Print("Cleaning temp debuff |cFFFF8080%s|r from |cFFFF0000%s", debuffName, mob.name)
					mob.debuffs[debuffName] = del(debuff)
				end
			end
			mob.temp = nil
			if (not next(mob.debuffs)) then
				--self:Print("Removing |cFFFF0000%s|r because of empty .debuffs", mob.name)
				mob.debuffs = del(mob.debuffs)
				self.mobs[guid] = del(self.mobs[guid])
			end
		end
	end
	--self:ResetInCombatForAll()
	--if (not next(self.combat)) then
	--	self:Cleanup()
	--end

	if (not UnitExists(target) or UnitIsDeadOrGhost(target) or not UnitCanAttack("player", target)) then
		return
	end

	local lookup = Utopia.lookup.debuffs
	if (not lookup) then
		return
	end

	local guid = UnitGUID(target)
	local mob = self.mobs[guid]
	local debuffs = mob and mob.debuffs

	for i = 1,100 do
		local name, rank, tex, count, debuffType, duration, endTime, caster, isStealable = UnitDebuff(target, i)
		if (name == nil) then
			break
		end

		local def = lookup[name]
		if (def) then
			if (not mob) then
				--self:Print("Creating temporary mob for |cFFFF0000%s", UnitName(target))
				mob = new()
				self.mobs[guid] = mob
				mob.name = UnitName(target)
				mob.guid = guid
				mob.debuffs = new()
				debuffs = mob.debuffs
			end

			local debuff = debuffs[name]
			if (not debuff) then
				-- Debuff exists on target that we've ignored in combat log

				--self:Print("- %s: New TEMP debuff: %s %s", UnitName(target), name, rank or "NoRank")
				debuff = new()
				debuffs[name] = debuff

				debuff.name = name
				debuff.rank = rank and tonumber(strmatch(rank, "(%d)"))
				debuff.source = self:ImpliedSourceGUID(caster, def)
				debuff.temp = true
				mob.temp = true
				debuff.stacks = count
			end
		end
	end
end

-- CheckTargetDebuffs
function engine:CheckTargetBuffs(target)
	if ((not UnitInRaid(target) and not UnitInParty(target)) or UnitIsDeadOrGhost(target)) then
		return
	end

	local lookup = Utopia.lookup.buffs
	if (not lookup) then
		return
	end

	local guid = UnitGUID(target)
	local plr = self.players[guid]
	if (not plr) then
		return
	end

	local old
	if (plr.buffs) then
		old = new()
		for buffname,info in pairs(plr.buffs) do
			old[buffname] = true
		end
	end

	for i = 1,100 do
		local name, rank, tex, count, debuffType, duration, endTime, caster, isStealable = UnitBuff(target, i)
		if (name == nil) then
			break
		end

		local def = lookup[name]
		if (def) then
			if (old and old[name]) then
				old[name] = nil
			end
			local buff = plr.buffs and plr.buffs[name]
			if (not buff) then
				--self:Print("- %s: New buff: %s %s", UnitName(target), name, rank or "NoRank")
				buff = new()
				if (not plr.buffs) then
					plr.buffs = new()
				end
				plr.buffs[name] = buff

				buff.name = name
				buff.rank = rank and tonumber(strmatch(rank, "(%d)"))
				if (name == def.alternate and def.alternateOffset) then
					buff.rank = buff.rank + def.alternateOffset
				end
				buff.source = self:ImpliedSourceGUID(caster, def)
				buff.stacks = count
			end
		end
	end
	if (old) then
		for name in pairs(old) do
			plr.buffs[name] = nil
		end
	end
end

-- ImpliedSourceGUID
if (isWoW3dot1) then
	function engine:ImpliedSourceGUID(caster, def)
		local guid = UnitGUID(caster)
		if (guid) then
			return guid
		end

		if (def.class) then
			if (Utopia.classcount[def.class] == 1 or (def.class == playerClass and Utopia.classcount[def.class] == 2)) then
				return self.someoneOfClass and self.someoneOfClass[def.class]
			end
		end
	end
else
	function engine:ImpliedSourceGUID(isMine, def)
		if (isMine) then
			return UnitGUID("player")
		elseif (def.class) then
			if (Utopia.classcount[def.class] == 1 or (def.class == playerClass and Utopia.classcount[def.class] == 2)) then
				return self.someoneOfClass and self.someoneOfClass[def.class]
			end
		end
	end
end

-- UNIT_NAME_UPDATE
function engine:UNIT_NAME_UPDATE(unit)
	local guid = UnitGUID(unit)
	local player = self.players[guid]
	if (player) then
		player.name = UnitFullName(unit)
	else
		local mob = self.mobs[guid]
		if (mob) then
			mob.name = UnitName(unit)
		end
	end
end

-- UNIT_PET
function engine:UNIT_PET(unit)
	--if (not UnitInParty(unit) and not UnitInRaid(unit)) then
	--	return
	--end

	local ownerGUID = UnitGUID(unit)
	if (not ownerGUID) then
		-- Probably zoning
		return 
	end

	local player = self.players[ownerGUID]
	if (not player) then
		-- We don't care about people outside the raid
		return
	end

	local petid
	if (unit == "player") then
		petid = "pet"
	else
		petid = unit:gsub("^(%a+)(%d+)$", "%1pet%2")
	end

	if (UnitExists(petid) and UnitCreatureFamily(petid)) then
		local petGUID = UnitGUID(petid)
		if (not self.players[petGUID]) then
			self:AddPlayer(petid, unit)
			player.pet = petGUID
			--self:Print("Added pet for %s called %s", UnitName(unit), UnitName(petid))
		end
	else
		local petGUID = player.pet
		if (petGUID) then
			player.pet = nil

			local obj = self.players[petGUID]
			if (obj) then
				obj.buffs = del(obj.buffs)
				obj.debuffs = del(obj.debuffs)
				--self:Print("Removed pet from %s", UnitName(unit))
				self.players[petGUID] = del(self.players[petGUID])
			end
		end
	end
end

-- AddPlayer
function engine:AddPlayer(unit, ownerid)
	local name = UnitFullName(unit)
	local guid = UnitGUID(unit)

	local n = self.oldRoster and self.oldRoster[guid]
	if (n) then
		self.oldRoster[guid] = nil
	else
		n = new()
	end
	self.players[guid] = n

	n.name = name
	n.guid = guid
	n.ownerid = ownerid

	return n
end

-- ScanRoster
function engine:ScanRoster()
	if (not self.players) then
		return
	end

	self.someoneOfClass = del(self.someoneOfClass)
	self.oldRoster = self.players
	self.players = new()
	self.scRogues = 0		-- BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------
	self.bfWarriors = 0

	for unit in Utopia:IterateRoster() do
		local player = self:AddPlayer(unit)

		----------------- BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------
		local _, class = UnitClass(unit)
		if (class == "ROGUE") then
			if (Utopia.talents and Utopia.talents[UnitFullName(unit)]) then
				if (Utopia:HasTalent(unit, GetSpellInfo(58413))) then
					self.scRogues = self.scRogues + 1
				end
			--else
			--	self.scRogues = self.scRogues + 1
			end
		elseif (class == "WARRIOR") then
			-- Dittoe for Warrior's Blood Frenzy
			if (Utopia.talents and Utopia.talents[UnitFullName(unit)]) then
				if (Utopia:HasTalent(unit, GetSpellInfo(29859))) then
					self.bfWarriors = self.bfWarriors + 1
				end
			--else
			--	self.scRogues = self.scRogues + 1
			end
		end
		-------------- END BUG FIX ----- Remove this when WoW includes Savage Combat in combat log -------------

		local petid
		if (unit == "player") then
			petid = "pet"
		else
			petid = unit:gsub("^(%a+)(%d+)$", "%1pet%2")
		end

		if (UnitExists(petid) and UnitCreatureFamily(petid)) then
			self:AddPlayer(petid, unit)
			player.pet = UnitGUID(petid)
		end

		if (class == playerClass and not UnitIsUnit(unit, "player")) then
			if (not self.someoneOfClass) then
				self.someoneOfClass = new()
			end
			self.someoneOfClass[class] = UnitGUID(unit)
		end
	end

	if (isWoW3dot1) then
		self:RegisterEvent("UNIT_AURA")
	else
		----------------- BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------
		if (self.scRogues > 0 or self.bfWarriors > 0 or Utopia.classcount.SHAMAN > 0) then
			self:RegisterEvent("UNIT_AURA")
		else
			self:UnregisterEvent("UNIT_AURA")
		end
		-------------- END BUG FIX ----- Remove this when WoW includes Savage Combat in combat log -------------
	end

	self.oldRoster = deepDel(self.oldRoster)
end

-- COMBAT_LOG_EVENT_UNFILTERED
local mask = COMBATLOG_OBJECT_AFFILIATION_PARTY + COMBATLOG_OBJECT_AFFILIATION_RAID + COMBATLOG_OBJECT_AFFILIATION_MINE
local dstMask = COMBATLOG_OBJECT_REACTION_HOSTILE + COMBATLOG_OBJECT_REACTION_NEUTRAL + COMBATLOG_OBJECT_CONTROL_NPC

function engine:COMBAT_LOG_EVENT_UNFILTERED(timestamp, event, ...)
	--for i = 1,select("#", ...) do
	--	local str = tostring(select(i, ...))
	--	if (str == "Savage Combat") then
	--		self:Print("|cFFFF0000%s in event %s", str, event)
	--		X = {timestamp, event, ...}
	--	end
	--end

	local ev = self.clEvents[event]
	if (ev) then
		ev(self, timestamp, event, ...)
	end
end

engine.clEvents = {}
-- SPELL_AURA_APPLIED
function engine.clEvents:SPELL_AURA_APPLIED(timestamp, event, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)
	if (band(srcFlags, mask) ~= 0) then
		if (auraType == "DEBUFF") then
			if (dstGUID ~= self.targetGUID and dstGUID ~= self.focusGUID) then
				if (band(dstFlags, dstMask) == 0) then
					return
				end
			end

			local player = self.players[srcGUID]
			if (not player) then
				-- Probably a hunter snake trap pet or something else minor
				return
			end

			----------------- BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------
			if (self.scRogues > 0 and spellSchool == SCHOOL_MASK_NATURE) then
				-- When a rogue with Savage Combat talent applies a poison, we remember who
				-- they are and the target they hit for next UNIT_AURA check... YUK!
				local _, class = UnitClass(srcName)
				if (class == "ROGUE") then
					if (SChaxCache[srcName]) then				-- Talent check cached for speed
						if (not self[savageCombat]) then
							self[savageCombat] = new()
						end
						--self:Print("Flagging |cFFFF8080Savage Combat|r potential for |cFFFFFF80%s|r on |cFFFF0000%s", srcName, dstName)
						self[savageCombat][dstGUID] = srcGUID
						return
					end
				end
			elseif (self.bfWarriors > 0 and spellSchool == SCHOOL_MASK_PHYSICAL) then
				if (spellName == GetSpellInfo(772) or spellId == 12721) then			-- Rend & Deep Wounds
					-- Same for Warriors who apply Rend
					local _, class = UnitClass(srcName)
					if (class == "WARRIOR") then
						if (SChaxCache[srcName]) then				-- Talent check cached for speed
							if (not self[bloodFrenzy]) then
								self[bloodFrenzy] = new()
							end
							--self:Print("Flagging |cFFFF8080Blood Frenzy|r potential for |cFFFFFF80%s|r on |cFFFF0000%s", srcName, dstName)
							self[bloodFrenzy][dstGUID] = srcGUID
							return
						end
					end
				end
			end
			------------- END BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------

			------- And remove this if they put Judgements of the Just and Master Poisoner in too.... ------
			local mapping = Utopia.mappingLookup[spellName]
			local mappedSpellName
			if (mapping) then
				--self:Print("Possible mapping for |cFFFF0000%s|r on |cFFFF8080%s|r from |cFFFFFF80%s", spellName, dstName, srcName)
				local _, class = UnitClass(srcName)
				if (mapping.class == class) then
					--self:Print("- Class matches")
					if (Utopia:HasTalent(srcName, mapping.requiredTalent)) then
						--self:Print("- Source has the |cFF80FF80%s|r talent", mapping.requiredTalent)
						mappedSpellName = spellName
						spellName = mapping.key
					--else
					--	self:Print("- Source does not have the |cFF80FF80%s|r talent", mapping.requiredTalent)
					end
				--else
				--	self:Print("- class does not match")			-- Won't happen
				end
			end

			if (not Utopia.lookup.debuffs[spellName]) then
				return
			end

			--self:Print("%s: |cFFFFFF80%s|r applies |cFFFF8080%s|r to |cFFFF0000%s|r - School = %s", event, srcName, spellName, dstName, tostring(spellSchool))
			local mob = self.mobs[dstGUID]
			if (not mob) then
				--self:Print("- New Mob: |cFFFF0000%s", dstName)
				mob = new()
				self.mobs[dstGUID] = mob

				mob.name = dstName
				mob.guid = dstGUID
				mob.debuffs = new()
			end

			if (mappedSpellName) then
				if (not self.mappings) then
					self.mappings = new()
				end
				if (not self.mappings[dstGUID]) then
					self.mappings[dstGUID] = new()
				end
				if (not self.mappings[dstGUID][srcGUID]) then
					self.mappings[dstGUID][srcGUID] = new()
				end
				--self:Print("- Added mapping for |cFFFF8080%s|r to |cFFFF8080%s", mappedSpellName, spellName)
				self.mappings[dstGUID][srcGUID][mappedSpellName] = spellName
			end
	
			local debuff = mob.debuffs[spellName]
			if (not debuff) then
				--self:Print("- New debuff for |cFFFF0000%s|r: |cFFFF8080%s", dstName, spellName)
				debuff = new()
				mob.debuffs[spellName] = debuff
	
				debuff.name = spellName
				debuff.id = spellId
				debuff.mapped = mappedSpellName
			end
	
			debuff.source = srcGUID
			--debuff.stacks = 1
			debuff.startTime = timestamp
	
			Utopia:OnDebuffApplied(dstGUID, spellName, debuff)
		else -- auraType = "BUFF"
			if (not db.collect.raid) then
				if (dstGUID ~= playerGUID or not db.collect.player) then
					return
				end
			end

			local player = self.players[srcGUID]
			if (not player) then
				return
			end

			self:AddBuff(timestamp, dstGUID, srcName, srcGUID, spellName, spellId)

			--local targetPlayer = self.players[dstGUID]
			--if (targetPlayer) then
			--	local buff = targetPlayer.buffs and targetPlayer.buffs[spellName]
			--	if (not buff) then
			--		--self:Print("- New debuff for |cFFFF0000%s|r: |cFFFF8080%s", dstName, spellName)
			--		buff = new()
			--		if (not targetPlayer.buffs) then
			--			targetPlayer.buffs = new()
			--		end
			--		targetPlayer.buffs[spellName] = buff
            --
			--		buff.name = spellName
			--		buff.id = spellId
			--	end
            --
			--	buff.source = srcGUID
			--	buff.startTime = timestamp
            --
			--	Utopia:OnBuffApplied(dstGUID, spellName, buff)
			--end
		end
	end
end

-- PostBuff
function engine:AddBuff(timestamp, dstGUID, srcName, srcGUID, spellName, spellId)
	local spellInfo = Utopia.lookup.buffs[spellName]
	if (not spellInfo) then
		return
	end

	local targetPlayer = self.players[dstGUID]
	if (targetPlayer) then
		local buff = targetPlayer.buffs and targetPlayer.buffs[spellName]
		if (not buff) then
			--self:Print("- New debuff for |cFFFF0000%s|r: |cFFFF8080%s", dstName, spellName)
			buff = new()
			if (not targetPlayer.buffs) then
				targetPlayer.buffs = new()
			end
			targetPlayer.buffs[spellName] = buff

			buff.name = spellName
			buff.id = spellId
		end

		buff.source = srcGUID
		buff.startTime = timestamp

		Utopia:OnBuffApplied(dstGUID, spellName, buff)

		if (spellInfo.trigger) then
			if (Utopia:HasTalent(srcName, spellInfo.trigger)) then
				--self:Print("Adding triggered spell: %s to %s", spellInfo.trigger, tostring(targetPlayer.name))
				self:AddBuff(timestamp, dstGUID, srcName, srcGUID, spellInfo.trigger)
			end
		end
	end
end

-- SPELL_AURA_APPLIED_DOSE
function engine.clEvents:SPELL_AURA_APPLIED_DOSE(timestamp, event, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)
	if (auraType == "DEBUFF") then
		if (not Utopia.lookup.debuffs[spellName]) then
			return
		end
		--self:Print("%s: |cFFFFFF80%s|r applies |cFFFF8080%s|r to |cFFFF0000%s|r", event, srcName, spellName, dstName)
		if (band(dstFlags, dstMask) == 0) then
			return
		end

		local mob = self.mobs[dstGUID]
		if (not mob) then
			-- self:Print("SPELL_AURA_APPLIED_DOSE(1) |cFFFF8080%s|r without a SPELL_AURA_APPLIED for %s", spellName or "nil", dstName or "nil")
			-- Probably targetted a mob being killed by someone outside group
			return
		end

		local debuff = mob.debuffs[spellName]
		if (debuff) then
			debuff.startTime = timestamp
			debuff.stacks = (debuff.stacks or 0) + 1
			--self:Print("- Stacks for |cFFFF0000%s's|r |cFFFF8080%s|r incremented to |cFF00FF00%d|r", dstName, spellName, debuff.stacks)
	
			Utopia:OnDebuffApplied(dstGUID, spellName, debuff, true)
		end
	else -- auraType = "BUFF"
		if (not db.collect.raid) then
			if (dstGUID ~= playerGUID or not db.collect.player) then
				return
			end
		end

		if (not Utopia.lookup.buffs[spellName]) then
			return
		end

		local targetPlayer = self.players[dstGUID]
		if (not targetPlayer) then
			return
		end

		local buff = targetPlayer.buffs and targetPlayer.buffs[spellName]
		if (buff) then
			buff.startTime = timestamp
			buff.stacks = (buff.stacks or 0) + 1

			Utopia:OnBuffApplied(dstGUID, spellName, buff, true)
		end
	end
end

-- SPELL_AURA_REMOVED
function engine.clEvents:SPELL_AURA_REMOVED(timestamp, event, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)
	if (band(srcFlags, mask) ~= 0) then
		if (auraType == "DEBUFF") then
			if (not Utopia.lookup.debuffs[spellName]) then
				if (not Utopia.mappingLookup[spellName]) then
					return
				end
			end
			if (dstGUID ~= self.targetGUID and dstGUID ~= self.focusGUID) then
				if (band(dstFlags, dstMask) == 0) then
					return
				end
			end

			if (srcGUID) then
				local destMapping = self.mappings and self.mappings[dstGUID]
				if (destMapping) then
					--self:Print("Checking mappings for |cFFFF0000%s", dstName)
					local srcMapping = destMapping[srcGUID]
					if (srcMapping) then
						--self:Print("- Found mappings from |cFFFFFF80%s", srcName)
						local spell = srcMapping[spellName]
						if (spell) then
							--self:Print("Mapped spell removed from |cFFFF0000%s|r: |cFFFF8080%s|r -> |cFFFF8080%s|r", dstName, spellName, spell)
							srcMapping[spellName] = nil
	
							if (not next(srcMapping)) then
								-- No more mappings from this class on this target, so we can remove the
								-- pseudo buff. This makes the assumption that there is only one mapping
								-- spell per class, which is safe given the dataset at this time
								--self:Print("- Removing mapping for |cFFFF8080%s|r to |cFFFF8080%s", spellName, spell)
								spellName = spell
	
								destMapping[srcGUID] = del(srcMapping)
								if (not next(destMapping)) then
									self.mappings[dstGUID] = del(destMapping)
									if (not next(self.mappings)) then
										self.mappings = del(self.mappings)
									end
								end
							--else
							--	self:Print("- Keeping mapping for |cFFFF8080%s|r", spell)
							end
						end
					end
				end
			end

			--self:Print("%s: |cFFFFFF80%s's|r debuff of |cFFFF8080%s|r removed from |cFFFF0000%s|r", event, srcName, spellName, dstName)
			local player = self.players[srcGUID]
			if (player) then
				local mob = self.mobs[dstGUID]
				if (mob) then
					local debuff = mob.debuffs[spellName]
					if (debuff) then
						Utopia:OnDebuffRemoved(dstGUID, spellName, debuff)
						mob.debuffs[spellName] = del(debuff)
					--else
						-- Probably ran out of combat and the mob's debuffs auto expired. This is ok.
					end
				--else
					-- Probably ran out of combat and the mob's debuffs auto expired. This is ok.
				end
			--else
				-- Probably debuff from a hunter snake trap pet expiring
			end
		else -- auraType == "BUFF"
			if (not db.collect.raid) then
				if (dstGUID ~= playerGUID or not db.collect.player) then
					return
				end
			end

			self:RemoveBuff(dstGUID, srcGUID, spellName)

			--if (not Utopia.lookup.buffs[spellName]) then
			--	return
			--end
            --
			--local player = self.players[srcGUID]
			--if (player) then
			--	local targetPlayer = self.players[dstGUID]
			--	if (targetPlayer) then
			--		local buff = targetPlayer.buffs and targetPlayer.buffs[spellName]
			--		if (buff) then
			--			Utopia:OnBuffRemoved(dstGUID, spellName, buff)
			--			targetPlayer.buffs[spellName] = nil
			--			if (not next(targetPlayer.buffs)) then
			--				targetPlayer.buffs = del(targetPlayer.buffs)
			--			end
			--		end
			--	end
			--end
		end
	end
end

-- RemoveBuff
function engine:RemoveBuff(dstGUID, srcGUID, spellName)
	local spellInfo = Utopia.lookup.buffs[spellName]
	if (not spellInfo) then
		return
	end

	local player = self.players[srcGUID]
	if (player) then
		local targetPlayer = self.players[dstGUID]
		if (targetPlayer) then
			local buff = targetPlayer.buffs and targetPlayer.buffs[spellName]
			if (buff) then
				Utopia:OnBuffRemoved(dstGUID, spellName, buff)

				if (spellInfo.trigger and targetPlayer.buffs) then
					for name, info in pairs(targetPlayer.buffs) do
						if (info.name == spellInfo.trigger) then
							if (info.srcGUID == srcGUID) then
self:Print("Removing triggered spell: %s from %s", spellInfo.trigger, tostring(targetPlayer.name))
								self:RemoveBuff(dstGUID, srcGUID, spellInfo.trigger)
							end
						end
					end
				end

				targetPlayer.buffs[spellName] = del(buff)
				if (not next(targetPlayer.buffs)) then
					targetPlayer.buffs = del(targetPlayer.buffs)
				end
			end
		end
	end
end

-- SPELL_AURA_REMOVED_DOSE
function engine.clEvents:SPELL_AURA_REMOVED_DOSE(timestamp, event, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)
	if (auraType == "DEBUFF") then
		if (not Utopia.lookup.debuffs[spellName]) then
			return
		end
		if (band(dstFlags, dstMask) == 0) then
			return
		end

		local mob = self.mobs[dstGUID]
		if (mob) then
			local debuff = mob.debuffs[spellName]
			if (debuff and debuff.stacks > 1) then
				debuff.stacks = debuff.stacks - 1
				Utopia:OnDebuffRemoved(dstGUID, spellName, debuff, true)
			end
		end
	else
		if (not db.collect.raid) then
			if (dstGUID ~= playerGUID or not db.collect.player) then
				return
			end
		end

		if (not Utopia.lookup.buffs[spellName]) then
			return
		end

		local targetPlayer = self.players[dstGUID]
		if (targetPlayer) then
			local buff = targetPlayer.debuffs[spellName]
			if (buff and buff.stacks > 1) then
				buff.stacks = buff.stacks - 1
				Utopia:OnDebuffRemoved(dstGUID, spellName, buff, true)
			end
		end
	end
end

-- SPELL_CREATE
function engine.clEvents:SPELL_SUMMON(timestamp, event, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, ...)
	if (band(srcFlags, mask) ~= 0) then
		if (not db.collect.raid) then
			if (srcGUID ~= playerGUID or not db.collect.player) then
				return
			end
		end

		local totem = Utopia.totems[spellName]
		if (totem) then
			if (not self.createdTotemGUIDs) then
				self.createdTotems = new()
				self.createdTotemGUIDs = new()
			end

			local t = new()
			t.srcGUID = srcGUID
			t.srcName = srcName
			t.timestamp = timestamp
			t.name = totem.totemBuffName or spellName
			t.totemName = spellName
			self.createdTotems[t.name] = t
			self.createdTotemGUIDs[dstGUID] = t
		end
	end
end

-- UNIT_DESTROYED
function engine.clEvents:UNIT_DIED(timestamp, event, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags)
	local found = self.createdTotemGUIDs and self.createdTotemGUIDs[dstGUID]
	if (found) then
		self.createdTotems[found.name] = nil
		self.createdTotemGUIDs[dstGUID] = nil
		del(found)

		if (not next(self.createdTotemGUIDs)) then
			self.createdTotems = del(self.createdTotems)
			self.createdTotemGUIDs = del(self.createdTotemGUIDs)
		end
	end
end

----------------- BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------
function engine:ApplyHackedAura(unit, spellName, onoff, ...)
	local dstGUID = UnitGUID(unit)

	if (onoff) then
		local srcGUID = self[spellName] and self[spellName][dstGUID]

		if (not self.has) then
			self.has = new()
		end
		if (not self.has[spellName]) then
			self.has[spellName] = new()
		end

		if (not self.has[spellName][dstGUID]) then
			self.has[spellName][dstGUID] = srcGUID

			local player = self.players[srcGUID]
			if (player) then
				local mob = self.mobs[dstGUID]
				if (not mob) then
					mob = new()
					self.mobs[dstGUID] = mob
	
					mob.name = UnitName(unit)
					mob.guid = dstGUID
					mob.debuffs = new()
				end
	
				local debuff = mob.debuffs[spellName]
				if (not debuff) then
					debuff = new()
					mob.debuffs[spellName] = debuff
				end
	
				debuff.name = spellName
				debuff.rank = onoff
				debuff.id = select(onoff, ...)
				debuff.source = srcGUID
				debuff.stacks = 1
	
				Utopia:OnDebuffApplied(dstGUID, spellName, debuff)
			end
		end
	else
		local mob = self.mobs[dstGUID]
		if (mob) then
			local debuff = mob.debuffs[spellName]
			if (debuff) then
				if (self.has and self.has[spellName] and self.has[spellName][dstGUID]) then
					self.has[spellName][dstGUID] = nil
				end

				Utopia:OnDebuffRemoved(dstGUID, spellName, debuff)
				mob.debuffs[spellName] = nil
			end
		end
	end
end
------------- END BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------

function engine:ApplyTotemBuff(unit, spellName, rank, ...)
	local dstGUID = UnitGUID(unit)

	if (rank) then
		local totem = self.createdTotems and self.createdTotems[spellName]
		local srcGUID = totem and totem.srcGUID

		if (not self.has) then
			self.has = new()
		end
		if (not self.has[spellName]) then
			self.has[spellName] = new()
		end

		if (not self.has[spellName][dstGUID]) then
			self.has[spellName][dstGUID] = srcGUID

			local player = self.players[srcGUID]
			if (player) then
				local mob = self.players[dstGUID]
				if (mob) then
					if (not mob.buffs) then
						mob.buffs = new()
					end

					local buff = mob.buffs[spellName]
					if (not buff) then
						buff = new()
						mob.buffs[spellName] = buff
					end
	
					buff.name = spellName
					buff.rank = rank
					--buff.id = select(rank, ...)
					buff.source = srcGUID
					--buff.stacks = 1

					Utopia:OnBuffApplied(dstGUID, spellName, buff)
				end
			end
		end
	else
		local mob = self.players[dstGUID]
		if (mob) then
			local buff = mob.buffs and mob.buffs[spellName]
			if (buff) then
				if (self.has and self.has[spellName] and self.has[spellName][dstGUID]) then
					self.has[spellName][dstGUID] = nil
				end

				Utopia:OnBuffRemoved(dstGUID, spellName, buff)
				mob.buffs[spellName] = nil
			end
		end
	end
end

-- UNIT_AURA
function engine:UNIT_AURA(unit)
	if (UnitCanAttack("player", unit)) then
		----------------- BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------
		if (self.scRogues > 0) then
			local name, rank = UnitDebuff(unit, savageCombat)
			rank = name and tonumber(rank:match("(%d)")) or nil
			self:ApplyHackedAura(unit, savageCombat, rank, 51682, 58413)
		end
		if (self.bfWarriors > 0) then
			local name, rank = UnitDebuff(unit, bloodFrenzy)
			rank = name and tonumber(rank:match("(%d)")) or nil
			self:ApplyHackedAura(unit, bloodFrenzy, rank, 29836, 29859)
		end
		------------- END BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------
	else
		if (Utopia.classcount.SHAMAN > 0) then
			local temp = copy(Utopia.allTotems)

			local i = 1
			while true do
				local name, rank = UnitBuff(unit, i)
				if (not name) then break end
				i = i + 1

				if (temp[name]) then
					self:ApplyTotemBuff(unit, name, tonumber(rank:match("(%d)")))
					temp[name] = nil
				end
			end

			for name in pairs(temp) do
				self:ApplyTotemBuff(unit, name)			-- No rank = Remove the buff
			end

			del(temp)
		end
	end
end

-- ResetInCombatForAll
function engine:ResetInCombatForAll()
	del(self.combat)
	self.combat = new()
	for unit in Utopia:IterateRoster() do
		self.combat[UnitGUID(unit)] = UnitAffectingCombat(unit)
	end
end

-- UNIT_FLAGS
function engine:UNIT_FLAGS(unit)
	if (UnitInParty(unit) or UnitPlayerOrPetInRaid(unit) or UnitPlayerOrPetInParty(unit)) then
		local guid = UnitGUID(unit)
		if (UnitAffectingCombat(unit)) then
			self.combat[guid] = true
		else
			if (self.combat[guid]) then
				self.combat[guid] = nil

				if (not next(self.combat) or UnitIsUnit(unit, "player")) then
					-- Nothing flagged as in-combat, but we double check to be sure
					-- for cases where ppl become IsVisible() and we don't see a
					-- UNIT_FLAGS event for them. Also checking this when I go out
					-- of combat to be doubly sure things don't go adrift over time
					self:ResetInCombatForAll()
					if (not next(self.combat)) then
						self:Cleanup()
					end
				end
			end
		end
	end
end

-- Cleanup
function engine:Cleanup()
	deepDel(self.mobs)
	self.mobs = new()
	deepDel(self.playerBuffs)
	self.playerBuffs = new()

	----------------- BUG FIX ----- Remove this when WoW includes Savage Combat/Blood Frenzy in combat log --------------
	self[savageCombat] = del(self[savageCombat])
	self[bloodFrenzy] = del(self[bloodFrenzy])
	self.has = deepDel(self.has)
	for k in pairs(SChaxCache) do
		SChaxCache[k] = nil
	end
	------------- END BUG FIX ----- Remove this when WoW includes Savage Combat in combat log --------------
	
	self.mappings = deepDel(self.mappings)
end

-- GetMobByUnit
function engine:GetMobByUnit(unit)
	return engine.mobs and engine.mobs[UnitGUID(unit)]
end

--@debug@
-- ShowFunc
-- Debug dump of all current debuffs on all targets
function engine:ShowFunc()
	if (self.mobs) then
		for guid, mob in pairs(self.mobs) do
			self:Print("|cFFFF0000%s|cFF808080 (%s)", mob.name, guid)

			if (next(mob.debuffs)) then
				for debuffName, debuff in pairs(mob.debuffs) do
					local GUID = debuff.source
					local player = engine.players[GUID]

					if (player) then
						self:Print("- |cFFFFFF80%s|r applied: |cFFFF8080%s|r (id=%d) (|cFF80FF80%s|r stacks)", player.name, debuffName, debuff.id, debuff.stacks)
					else
						self:Print("- |cFFFF00FFUnknown|r applied %s (id=%d) (|cFF80FF80%s|r stacks)", debuffName, debuff.id, debuff.stacks)
					end
				end
			else
				self:Print("- No debuffs")
			end
		end
	end
end

local function UE()
	engine:ShowFunc()
end

SlashCmdList["UE"] = UE
SLASH_UE1 = "/ue"
--@no-debug@
